# Preprocess input for use in Order-Score
# Convert to TSV; ensure data numerically coded; subset genes of interest; remove header; 
# input: banjo-formatted text file of discretized gene expression (n x M)
# output: order-search formatted text file of discretized gene expression (n x M); list of node names
library(stringr)


# setup -------------------------------------------------------------------
analysisDir <- "./analysis/order-search/"
analysis <- "GSE141593_piz_ko_mice_gene_expr_suppgenes_n=215_log2fc=2"
expr_file <- paste0("./analysis/banjo/input/", analysis, ".txt")
gene_list <- paste0(analysisDir, "/input/order_search_genes.txt")


# load data ---------------------------------------------------------------
expr_data <- read.table(expr_file, sep = "\t", header = TRUE)
genes <- read.table(gene_list, sep = "\n", header = FALSE)$V1 |> 
  str_to_sentence()

head(expr_data, c(4,5))


# subset ------------------------------------------------------------------
all(genes %in% colnames(expr_data)) # false
# genes[!genes %in% colnames(expr_data)] # which are missing

included_genes <- genes[genes %in% colnames(expr_data)]
expr_data <- expr_data[,c("pizMutation", included_genes)]

write.table(expr_data, file = paste0(analysisDir, "/input/", analysis, "_representative.txt"),
            row.names = FALSE, col.names = FALSE, quote = FALSE)

write.table(colnames(expr_data), file = paste0(analysisDir, "/input/", analysis, "_nodes.txt"),
            row.names = FALSE, col.names = FALSE, quote = FALSE)